import 'dart:async';
import 'dart:convert';
import 'storage/html.dart' if (dart.library.io) 'storage/io.dart';

class GetStorage {
  static final Map<String, GetStorage> _sync = Map();

  StreamSubscription<Map<String, dynamic>> listen(
          void onData(Map<String, dynamic> event),
          {Function onError,
          void onDone(),
          bool cancelOnError}) =>
      stream.listen(onData,
          onError: onError, onDone: onDone, cancelOnError: cancelOnError);

  factory GetStorage(
      [String container = "GetStorage",
      String path,
      Map<String, dynamic> initialData]) {
    if (_sync.containsKey(container)) {
      return _sync[container];
    } else {
      final instance = GetStorage._internal(container, path, initialData);
      _sync[container] = instance;
      return instance;
    }
  }

  void dispose() {
    _concrete?.dispose();
  }

  GetStorage._internal(String key,
      [String path, Map<String, dynamic> initialData]) {
    _concrete = StorageImpl(key, path);
    _initialData = initialData;

    isReady = Future<bool>(() async {
      await this._init();
      return true;
    });
  }

  Future<void> _init() async {
    try {
      await _concrete.init(_initialData);
    } on Error catch (err) {
      throw err;
    }
  }

  dynamic read(String key) {
    return _concrete.read(key);
  }

  Future<void> write(
    String key,
    value, [
    Object toEncodable(Object nonEncodable),
  ]) async {
    final _encoded =
        json.encode(toEncodable != null ? toEncodable(value) : value);
    await _concrete.write(key, json.decode(_encoded));

    return _attemptFlush();
  }

  Future<void> remove(String key) async {
    await _concrete.remove(key);
    return _attemptFlush();
  }

  Future<void> clear() async {
    await _concrete.clear();
    return _attemptFlush();
  }

  Future<void> _attemptFlush() async {
    if (_lock != null) {
      await _lock;
    }
    _lock = _flush();
    return _lock;
  }

  Future<void> _flush() async {
    try {
      await _concrete.flush();
    } catch (e) {
      rethrow;
    }
    return;
  }

  StorageImpl _concrete;

  Future<bool> isReady;

  Future<void> _lock;

  Stream<Map<String, dynamic>> get stream => _concrete.stream;

  Map<String, dynamic> _initialData;
}
